from cvmodelz.models.pretrained.base import PretrainedModelMixin
from cvmodelz.models.pretrained.inception import InceptionV3
from cvmodelz.models.pretrained.resnet import ResNet101
from cvmodelz.models.pretrained.resnet import ResNet152
from cvmodelz.models.pretrained.resnet import ResNet35
from cvmodelz.models.pretrained.resnet import ResNet50
from cvmodelz.models.pretrained.vgg import VGG16
from cvmodelz.models.pretrained.vgg import VGG19


__all__ = [
	"PretrainedModelMixin",

	"VGG16",
	"VGG19",

	"ResNet35",
	"ResNet50",
	"ResNet101",
	"ResNet152",

	"InceptionV3",
]



import abc
import chainer
import io

from chainer import functions as F
from chainer.serializers import npz
from typing import Callable
from typing import Dict

from cvmodelz.models.base import BaseModel

class Classifier(chainer.Chain):

	def __init__(self, model: BaseModel, *,
		layer_name: str = None,
		loss_func: Callable = F.softmax_cross_entropy,
		only_head: bool = False,
		):
		super().__init__()
		self._only_head = only_head
		self.layer_name = layer_name or model.clf_layer_name
		self.loss_func = loss_func

		with self.init_scope():
			self.setup(model)

		if only_head:
			self.enable_only_head()

	def setup(self, model: BaseModel) -> None:
		self.model = model

	def report(self, **values) -> None:
		chainer.report(values, self)

	def enable_only_head(self) -> None:
		self.model.disable_update()
		self.model.clf_layer.enable_update()

	@property
	def n_classes(self) -> int:
		return self.model.clf_layer.W.shape[0]

	def save(self, weights_file):
		npz.save_npz(weights_file, self)

	def load(self, weights_file: str, n_classes: int, *, finetune: bool = False, **kwargs) -> None:
		""" Loading a classifier has following use cases:

			(0) No loading.
				- All weights are initilized randomly.

			(1) Loading from default pre-trained weights
				- The weights are loaded directly to
				the model.
				- Any additional not model-related
				layer will be initialized randomly.

			(2) Loading from a saved classifier.
				- All weights are loaded as-it-is from
				the given file.
		"""

		try:
			# Case (2)
			self.load_classifier(weights_file)

		except KeyError as e:
			pass

		# Case (1)
		self.load_model(weights_file, n_classes=n_classes, finetune=finetune, **kwargs)

		# else:
		# 	# Case (0)
		# 	pass

	def load_classifier(self, weights_file: str):

		if isinstance(weights_file, io.BufferedIOBase):
			assert not weights_file.closed, "The weights file was already closed!"
			weights_file.seek(0)

		npz.load_npz(weights_file, self, strict=True)

	def get_model_loader(self, finetune: bool = False, model: BaseModel = None):
		model = model or self.model
		if finetune:
			return model.load_for_finetune
		else:
			return model.load_for_inference


	def load_model(self, weights_file: str, n_classes: int, *, finetune: bool = False, **kwargs):
		model_loader = self.get_model_loader(finetune=finetune, model=self.model)
		kwargs["strict"] = kwargs.get("strict", True)
		model_loader(weights=weights_file, n_classes=n_classes, **kwargs)

	@property
	def feat_size(self) -> int:
		return self.model.meta.feature_size

	@property
	def output_size(self) -> int:
		return self.feat_size

	def loss(self, pred: chainer.Variable, y: chainer.Variable) -> chainer.Variable:
		return self.model.loss(pred, y, loss_func=self.loss_func)

	def evaluations(self, pred: chainer.Variable, y: chainer.Variable) -> Dict[str, chainer.Variable]:
		return dict(accuracy=self.model.accuracy(pred, y))

	def forward(self, X: chainer.Variable, y: chainer.Variable) -> chainer.Variable:
		pred = self.model(X, layer_name=self.layer_name)

		loss = self.loss(pred, y)
		evaluations = self.evaluations(pred, y)

		self.report(loss=loss, **evaluations)
		return loss





import abc
import pyaml


from chainer import links as L
from chainercv2.models import inceptionv3 as cv2inceptionv3
from chainercv2.models import resnet as cv2resnet
from collections import OrderedDict

from cvmodelz.models import pretrained
from cvmodelz.models.wrapper import ModelWrapper

class ModelFactory(abc.ABC):

	supported = OrderedDict(
		chainer=(
			L.ResNet50Layers,
			L.ResNet101Layers,
			L.ResNet152Layers,
			# L.VGG16Layers,
			# L.VGG19Layers,
		),

		chainercv=(
			# todo: chainercv.links.models.ssd
		),

		chainercv2=(
			cv2resnet.resnet50,
			# cv2resnet.resnet50b,

			cv2inceptionv3.inceptionv3,
		),

		cvmodelz=(
			pretrained.VGG16,
			pretrained.VGG19,

			pretrained.ResNet35,
			pretrained.ResNet50,
			pretrained.ResNet101,
			pretrained.ResNet152,

			pretrained.InceptionV3,
		),
	)

	@abc.abstractmethod
	def __init__(self):
		raise NotImplementedError("instance creation is not supported!")

	@classmethod
	def new(cls, model_type, **kwargs):

		key, cls_name = model_type.split(".")

		for model_cls in cls.supported[key]:
			if model_cls.__name__ == cls_name:
				break
		else:
			raise ValueError(f"Could not find {model_type}!")

		if model_cls in cls.supported["chainercv2"]:
			n_classes = kwargs.pop("n_classes", 1000)
			pretrained = kwargs.pop("pretrained_model", None) == "auto"
			model = model_cls(classes=n_classes, pretrained=pretrained)
			kwargs["model"] = model
			model_cls = ModelWrapper

		return model_cls(**kwargs)


	@classmethod
	def _check(cls, model, key):
		return isinstance(model, cls.supported[key])

	@classmethod
	def is_chainer_model(cls, model):
		return cls._check(model, "chainer")

	@classmethod
	def is_cv_model(cls, model):
		return cls._check(model, "chainercv")

	@classmethod
	def is_cv2_model(cls, model):
		return cls._check(model, "chainercv2")

	@classmethod
	def is_cvmodelz_model(cls, model):
		return cls._check(model, "cvmodelz")

	@classmethod
	def get_all_models(cls, key=None):
		if key is not None:
			return [f"{key}.{model_cls.__name__}" for model_cls in cls.supported[key]]

		return cls.get_models(cls.supported.keys())

	@classmethod
	def get_models(cls, keys=None):
		if keys is None:
			keys = cls.supported.keys()

		res = []
		for key in keys:
			res += cls.get_all_models(key)

		return res


if __name__ == '__main__':
	print(pyaml.dump(dict(Models=ModelFactory.get_all_models()), indent=2))



from cvmodelz.models.base import BaseModel
from cvmodelz.models.factory import ModelFactory
from cvmodelz.models.pretrained import InceptionV3
from cvmodelz.models.pretrained import ResNet101
from cvmodelz.models.pretrained import ResNet152
from cvmodelz.models.pretrained import ResNet35
from cvmodelz.models.pretrained import ResNet50
from cvmodelz.models.pretrained import VGG16
from cvmodelz.models.pretrained import VGG19

__all__ = [
	"BaseModel",
	"ModelFactory",

	"InceptionV3",

	"ResNet50",
	"ResNet35",
	"ResNet101",
	"ResNet152",

	"VGG16",
	"VGG19",
]



from chainer import links as L
from chainer.links.model.vision.vgg import prepare
from chainer.links.model.vision.vgg import _max_pooling_2d

from cvmodelz.models.meta_info import ModelInfo
from cvmodelz.models.pretrained.base import PretrainedModelMixin

def _vgg_meta(final_conv_layer):
	return ModelInfo(
		name="VGG",
		input_size=224,
		feature_size=4096,
		n_conv_maps=512,

		conv_map_layer=final_conv_layer,
		feature_layer="fc7",

		classifier_layers=["fc6", "fc7", "fc8"],

		prepare_func=prepare,
	)

class BaseVGG(PretrainedModelMixin):
	def __init__(self, *args, **kwargs):
		kwargs["pooling"] = kwargs.get("pooling", _max_pooling_2d)
		super().__init__(*args, **kwargs)

	@property
	def functions(self):
		return super().functions

	def init_model_info(self):
		self.meta = _vgg_meta(self.final_conv_layer)

class VGG19(BaseVGG, L.VGG19Layers):
	final_conv_layer = "conv5_4"


class VGG16(BaseVGG, L.VGG16Layers):
	final_conv_layer = "conv5_3"




import chainer

from chainer import functions as F
from collections import OrderedDict

from cvmodelz.models.base import BaseModel
from cvmodelz.models.meta_info import ModelInfo



class ModelWrapper(BaseModel):
	"""
		This class is designed to wrap around chainercv2 models
		and provide the loading API of the BaseModel class.
		The wrapped model is stored under self.wrapped
	"""

	def __init__(self, model: chainer.Chain, *args, **kwargs):
		super().__init__(*args, **kwargs)

		name = model.__class__.__name__
		self.__class__.__name__ = name
		self.meta.name = name

		if hasattr(model, "meta"):
			self.meta = model.meta

		with self.init_scope():
			self.wrapped = model
			delattr(self.wrapped.features, "final_pool")

		self.meta.feature_size = self.clf_layer.W.shape[-1]

	def init_model_info(self):
		self.meta = ModelInfo(
			classifier_layers=("output/fc",),
			conv_map_layer="features",
			feature_layer="pool",
		)

	@property
	def model_instance(self) -> chainer.Chain:
		return self.wrapped

	@property
	def functions(self) -> OrderedDict:

		links = [
			("features", [self.wrapped.features]),
			("pool", [self.pool]),
			("output/fc", [self.wrapped.output.fc]),
		]

		return OrderedDict(links)

	def load(self, *args, path="", **kwargs):
		paths = [path, f"{path}wrapped/"]
		for _path in paths:
			try:
				return super().load(*args, path=_path, **kwargs)
			except KeyError as e:
				pass

		raise RuntimeError(f"tried to load weights with paths {paths}, but did not succeeed")

	def forward(self, X, layer_name=None):
		if layer_name is None:
			res = self.wrapped(X)

		elif layer_name == self.meta.conv_map_layer:
			res = self.wrapped.features(X)

		elif layer_name == self.meta.feature_layer:
			conv = self.wrapped.features(X)
			res = self.pool(conv)

		elif layer_name == self.clf_layer_name:
			conv = self.wrapped.features(X)
			feat = self.pool(conv)
			res = self.wrapped.output(feat)

		else:
			raise ValueError(f"Dont know how to compute \"{layer_name}\"!")

		return res




#!/usr/bin/env python
if __name__ != '__main__': raise Exception("Do not import me!")

from cvargparse import Arg
from cvargparse import BaseParser

from cvmodelz import utils
from cvmodelz.models import ModelFactory

def main(args):

	model = ModelFactory.new(args.model_type, input_size=args.input_size)
	utils.print_model_info(model)

parser = BaseParser()

parser.add_args([
	Arg("model_type", choices=ModelFactory.get_all_models()),

	Arg("--input_size", "-size", type=int, default=None)
])

main(parser.parse_args())



from functools import reduce
import chainer
import numpy as np
import sys
import logging

from functools import partial
from tabulate import tabulate

def get_attr_from_path(obj, path, *, sep="/"):
	def getter(o, attr):
		return
	getter = lambda o, attr: (getattr(o, attr) if attr else o)
	return reduce(getter, path.split(sep), obj)

def _get_activation_shapes(model, input_size, input_var, batch_size=2, n_channels=3):
	assert hasattr(model, "functions"), \
		"Model should have functions defined!"

	if input_var is None:
		input_shape = (batch_size, n_channels, input_size, input_size)
		x = model.xp.zeros(input_shape, dtype=model.xp.float32)
	else:
		x = input_var

	res = [("Input", x.shape)]
	with chainer.no_backprop_mode(), chainer.using_config("train", False):
		for name, link in model.functions.items():
			in_shape = str(x.shape)
			out_shapes = []
			for func in link:
				x = func(x)
				out_shapes.append(str(x.shape))
			logging.debug("\t".join(map(str, (name, in_shape, out_shapes))))
			res.append((name, in_shape, " -> ".join(out_shapes)))
	return res

def print_model_info(model, file=sys.stdout, input_size=None, input_var=None):
	_print = partial(print, file=file)
	name = getattr(model, "name", None)
	if name is None:
		name = model.__class__.__name__

	rows = []
	default_size = 224
	if hasattr(model, "meta"):
		_print(model.meta)
		default_size = model.meta.input_size
		rows.append(("Default input size", f"{default_size}"))

		feature_size = model.meta.feature_size
		rows.append(("Feature size", f"{feature_size}"))

		n_conv_maps = model.meta.n_conv_maps
		rows.append(("# of conv maps (last layer)", f"{n_conv_maps}"))

	n_weights = model.count_params()
	rows.append(("# of parameters", f"{n_weights:,d}"))

	n_params = len(list(model.params()))
	rows.append(("# of trainables", f"{n_params:,d}"))

	n_layers = len(list(model.links()))
	rows.append(("# of layers", f"{n_layers:,d}"))

	_print(f"Printing some information about \"{name}\" model")
	_print(tabulate(rows, tablefmt="fancy_grid"))

	shapes = _get_activation_shapes(model, input_size or default_size, input_var)
	_print("In/Out activation shapes:")
	_print(tabulate(shapes,
		headers=["Link name", "Input", "Output"],
		tablefmt="fancy_grid"))




import abc
import chainer

from chainer import functions as F
from typing import Callable
from typing import Dict

from cvmodelz.classifiers.base import Classifier
from cvmodelz.models.base import BaseModel



class SeparateModelClassifier(Classifier):
	"""
		Abstract Classifier, that holds two separate models.
		The user has to define, how these models operate on the
		input data. Hence, the forward method is abstract!
	"""

	@abc.abstractmethod
	def forward(self, *args, **kwargs) -> chainer.Variable:
		super().forward(*args, **kwargs)

	def setup(self, model: BaseModel) -> None:
		super().setup(model)

		self.separate_model = self.model.copy(mode="copy")

	def load_model(self, weights_file: str, n_classes: int, *, finetune: bool = False, **kwargs) -> None:
		for model in [self.model, self.separate_model]:
			model_loader = self.get_model_loader(finetune=finetune, model=model)
			kwargs["strict"] = kwargs.get("strict", True)
			model_loader(weights=weights_file, n_classes=n_classes, **kwargs)

	def enable_only_head(self) -> None:
		super().enable_only_head()
		self.separate_model.disable_update()
		self.separate_model.fc.enable_update()


class MeanModelClassifier(SeparateModelClassifier):

	def evaluations(self, pred0: chainer.Variable, pred1: chainer.Variable, y: chainer.Variable) -> Dict[str, chainer.Variable]:
		accu0 = self.model.accuracy(pred0, y)
		accu1 = self.separate_model.accuracy(pred1, y)

		mean_pred = (F.softmax(pred0) + F.softmax(pred1)) / 2

		accu = self.model.accuracy(mean_pred, y)

		return dict(
			accu0=accu0,
			accu1=accu1,
			accuracy=accu,
		)

	def forward(self, X: chainer.Variable, y: chainer.Variable) -> chainer.Variable:

		pred0 = self.model(X, layer_name=self.layer_name)
		pred1 = self.separate_model(X, layer_name=self.layer_name)

		loss0, loss1 = self.loss(pred0, y), self.loss(pred1, y)
		loss = (loss0 + loss1) / 2

		evaluations = self.evaluations(pred0, pred1, y)

		self.report(
			loss0=loss0,
			loss1=loss1,
			loss=loss,
			**evaluations
		)
		return loss






import abc
import chainer
import io
import numpy as np

from chainer import functions as F
from chainer import links as L
from chainer.initializers import HeNormal
from chainer.serializers import npz
from chainer_addons.links.pooling import PoolingType # TODO: replace this!
from collections import OrderedDict
from typing import Callable

from cvmodelz import utils
from cvmodelz.models.meta_info import ModelInfo

class BaseModel(abc.ABC, chainer.Chain):

	def __init__(self, pooling: Callable = PoolingType.G_AVG,
		input_size=None, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.init_model_info()

		if isinstance(pooling, (PoolingType, str)):
			pooling = PoolingType.new(pooling)

		with self.init_scope():
			self.pool = pooling

		if input_size is not None:
			self.meta.input_size = input_size


	def init_model_info(self):
		self.meta = ModelInfo()

	@abc.abstractmethod
	def forward(self, *args, **kwargs):
		return super().forward(*args, **kwargs)

	@abc.abstractproperty
	def functions(self) -> OrderedDict:
		return super().functions

	@abc.abstractproperty
	def model_instance(self) -> chainer.Chain:
		raise NotImplementedError()

	@property
	def clf_layer_name(self) -> str:
		return self.meta.classifier_layers[-1]

	@property
	def clf_layer(self) -> chainer.Link:
		return utils.get_attr_from_path(self.model_instance, self.clf_layer_name)

	def loss(self, pred, gt, loss_func=F.softmax_cross_entropy):
		return loss_func(pred, gt)

	def accuracy(self, pred, gt):
		return F.accuracy(pred, gt)

	def reinitialize_clf(self, n_classes, feat_size=None, initializer=None):

		if initializer is None or not callable(initializer):
			initializer = HeNormal(scale=1.0)

		clf_layer = self.clf_layer

		assert isinstance(clf_layer, L.Linear)

		w_shape = (n_classes, feat_size or clf_layer.W.shape[1])
		dtype = clf_layer.W.dtype

		clf_layer.out_size, clf_layer.in_size = w_shape
		clf_layer.W.data = np.zeros(w_shape, dtype=dtype)
		clf_layer.b.data = np.zeros(w_shape[0], dtype=dtype)

		initializer(clf_layer.W.data)

	def load_for_finetune(self, weights, n_classes, *, path="", strict=False, headless=False, **kwargs):
		"""
			The weights should be pre-trained on a bigger
			dataset (eg. ImageNet). The classification layer is
			reinitialized after all other weights are loaded
		"""
		self.load(weights, path=path, strict=strict, headless=headless)
		self.reinitialize_clf(n_classes, **kwargs)

	def load_for_inference(self, weights, n_classes, *, path="", strict=False, headless=False, **kwargs):
		"""
			In this use case we are loading already fine-tuned
			weights. This means, we need to reinitialize the
			classification layer first and then load the weights.
		"""
		self.reinitialize_clf(n_classes, **kwargs)
		self.load(weights, path=path, strict=strict, headless=headless)

	def load(self, weights, *, path="", strict=False, headless=False):
		if weights in [None, "auto"]:
			return

		ignore_names = None
		if headless:
			ignore_names = lambda name: name.startswith(path + self.clf_layer_name)

		if isinstance(weights, io.BufferedIOBase):
			assert not weights.closed, "The weights file was already closed!"
			weights.seek(0)

		npz.load_npz(weights, self.model_instance,
			path=path, strict=strict, ignore_names=ignore_names)

	def save(self, path, *args, **kwargs):
		npz.save_npz(path, self, *args, **kwargs)



import pyaml

from dataclasses import dataclass
from typing import Tuple
from typing import Callable


@dataclass
class ModelInfo(object):

	name:                       str         = "GenericModel"
	input_size:                 int         = 224
	feature_size:               int         = 2048
	n_conv_maps:                int         = 2048

	conv_map_layer:             str         = "conv"
	feature_layer:              str         = "fc"

	classifier_layers:          Tuple[str]  = ("fc",)

	prepare_func:               Callable    = None

	def __str__(self):
		obj = dict(ModelInfo=self.__dict__)
		return pyaml.dump(obj, sort_dicts=False, )


if __name__ == '__main__':
	print(ModelInfo())



import chainer
import chainer.functions as F
import chainer.links as L
import numpy as np

from chainer_addons.links.pooling import PoolingType # TODO: replace this!
from chainercv.transforms import resize
from chainercv.transforms import scale
from collections import OrderedDict
from collections.abc import Iterable


from cvmodelz.models.meta_info import ModelInfo
from cvmodelz.models.pretrained.base import PretrainedModelMixin
from cvmodelz.models.pretrained.inception import blocks
from cvmodelz.models.pretrained.inception import link_mappings


def _assign(name, param, data):
	assert data.shape == param.shape, \
		"\"{}\" does not match the shape: {} != {}!".format(
			name, data.shape, param.shape)
	if isinstance(param, chainer.variable.Parameter):
		param.data[:] = data
	else:
		param[:] = data

def _assign_batch_norm(name, link, beta, avg_mean, avg_var):
	_assign(name, link.beta, beta)
	_assign(name, link.avg_mean, avg_mean)
	_assign(name, link.avg_var, avg_var)



"""
We need this to "extract" pretrained_model argument,
otherwise it would be passed to the constructor of the
chainer.Chain class, where it raises an error
"""
class InceptionV3Layers(chainer.Chain):


	def __init__(self, pretrained_model=None, aux_logits=False, *args, **kwargs):
		self.aux_logits = aux_logits
		super().__init__(*args, **kwargs)


class InceptionV3(PretrainedModelMixin, InceptionV3Layers):

	def init_model_info(self):
		self.meta = ModelInfo(
			name="InceptionV3",
			input_size=299,
			feature_size=2048,
			n_conv_maps=2048,

			conv_map_layer="mixed10",
			feature_layer="pool",

			classifier_layers=["fc"],

			prepare_func=self.prepare
		)

	@property
	def functions(self):
		super().functions

	def prepare(self, x, size=None, *, swap_channels=False, keep_ratio=True):
		size = size or self.meta.input_size

		# [0 .. 255] -> [0 .. 1]
		x = x.astype(np.float32) / 255

		# HWC -> CHW channel order
		x = x.transpose(2,0,1)

		if keep_ratio:
			if isinstance(size, Iterable):
				size = min(size)
			# scale the smallest side to <size>
			x = scale(x, size=size)
		else:
			if isinstance(size, int):
				size = (size, size)

			# resize the image to  side to <size>x<size>
			x = resize(x, size)

		if swap_channels:
			x = x[::-1]
		return x


	def forward(self, x, layer_name='fc'):
		aux_logit = None
		for key, funcs in self.functions.items():
			for func in funcs:
				x = func(x)

				if chainer.config.train and self.aux_logits and func == self.mixed07:
					aux_logit = self.aux(x)

			if key == layer_name:
				return x if aux_logit is None else (x, aux_logit)

	@property
	def functions(self):

		names = ["mixed{:02d}".format(i) for i in range(11)]
		body = [(name, [getattr(self, name)]) for name in names]
		links = [
			("head", [self.head]),
		] + body + [
			("pool", [self.pool]),
			("fc", [self.fc]),
		]

		return OrderedDict(links)

	def extract(self, x):
		x = self.head(x)
		x = self.mixed00(x)
		x = self.mixed01(x)
		x = self.mixed02(x)
		x = self.mixed03(x)
		x = self.mixed04(x)
		x = self.mixed05(x)
		x = self.mixed06(x)
		x = self.mixed07(x)
		x = self.mixed08(x)
		x = self.mixed09(x)
		x = self.mixed10(x)

		return self.pool(x)


	def load(self, weights, *args, **kwargs) -> None:
		if isinstance(weights, str) and weights.endswith(".h5"):
			return self._load_from_keras(weights)
		elif isinstance(weights, str) and weights.endswith(".ckpt.npz"):
			return self._load_from_ckpt_weights(weights)

		return super().load(weights, *args, **kwargs)

	def init_extra_layers(self, n_classes) -> None:
		# input 3 x 299 x 299
		self.head = blocks.InceptionHead()
		# output 192 x 35 x 35


		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 192 x 35 x 35
		self.mixed00 = blocks.Inception1(insize=192, sizes=[48, 64, 96], outputs=[64, 64, 96, 32], **pool_args)
		# input 256 x 35 x 35
		self.mixed01 = blocks.Inception1(insize=256, sizes=[48, 64, 96], outputs=[64, 64, 96, 64], **pool_args)
		# input 288 x 35 x 35
		self.mixed02 = blocks.Inception1(insize=288, sizes=[48, 64, 96], outputs=[64, 64, 96, 64], **pool_args)

		pool_args = dict(pool_type=PoolingType.MAX, ksize=3, stride=2, pad=0)
		# input 288 x 35 x 35
		self.mixed03 = blocks.Inception2(288, sizes=[64, 96], outputs=[384, 96], **pool_args)

		# input 768 x 17 x 17
		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 768 x 17 x 17
		self.mixed04 = blocks.Inception3(768, sizes=[128] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed05 = blocks.Inception3(768, sizes=[160] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed06 = blocks.Inception3(768, sizes=[160] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed07 = blocks.Inception3(768, sizes=[192] * 6, outputs=[192] * 4, **pool_args)

		self.aux = blocks.AuxilaryClassifier(n_classes) if self.aux_logits else F.identity

		pool_args = dict(pool_type=PoolingType.MAX, ksize=3, stride=2, pad=0)
		# input 768 x 17 x 17
		self.mixed08 = blocks.Inception4(768, sizes=[192] * 4, outputs=[320, 192], **pool_args)
		# output 1280 x 8 x 8

		# input 1280 x 8 x 8
		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 1280 x 8 x 8
		self.mixed09 = blocks.Inception5(1280, sizes=[384, 448, 384], outputs=[320, 384, 384, 192], **pool_args)
		# input 2048 x 8 x 8
		self.mixed10 = blocks.Inception5(2048, sizes=[384, 448, 384], outputs=[320, 384, 384, 192], **pool_args)

		# input 2048 x 8 x 8
		# global average pooling
		# output 2048 x 1 x 1

		# the final fc layer is initilized by PretrainedModelMixin
		super().init_extra_layers(n_classes)

	def loss(self, pred, gt, loss_func=F.softmax_cross_entropy, alpha=0.4):
		if isinstance(pred, tuple):
			pred0, aux_pred = pred
		else:
			pred0, aux_pred = pred, None

		loss = loss_func(pred0, gt)
		if aux_pred is None:
			return loss
		else:
			aux_loss = loss_func(aux_pred, gt)
			return (1-alpha) * loss + alpha * aux_loss

	def accuracy(self, pred, gt):
		if isinstance(pred, tuple):
			pred0, aux_pred = pred
		else:
			pred0, aux_pred = pred, None
		return F.accuracy(pred0, gt)

	def _load_from_ckpt_weights(self, weights):
		content = np.load(weights)
		for name, link in self.namedlinks(skipself=True):
			if name not in link_mappings.chainer_to_tf_ckpt:
				continue

			ckpt_key = link_mappings.chainer_to_tf_ckpt[name]

			if isinstance(link, L.Convolution2D):
				W = content["{}/weights".format(ckpt_key)]
				W = W.transpose(3,2,0,1)
				_assign(name, link.W, W)

			elif isinstance(link, L.BatchNormalization):
				beta = content["{}/beta".format(ckpt_key)]
				avg_mean = content["{}/moving_mean".format(ckpt_key)]
				avg_var = content["{}/moving_variance".format(ckpt_key)]
				_assign_batch_norm(name, link, beta, avg_mean, avg_var)

			elif isinstance(link, L.Linear):
				W = content["{}/weights".format(ckpt_key)]
				W = W.transpose(3,2,0,1).squeeze()
				b = content["{}/biases".format(ckpt_key)]
				_assign(name, link.W, W)
				_assign(name, link.b, b)

			else:
				raise ValueError("Unkown link type: {}!".format(type(link)))

	def _load_from_keras(self, weights):
		import h5py
		with h5py.File(weights, "r") as f:
			if "model_weights" in f:
				f = f["model_weights"]
			for name, link in self.namedlinks(skipself=True):
				if name not in link_mappings.chainer_to_keras: continue
				keras_key = link_mappings.chainer_to_keras[name]

				if isinstance(link, L.Convolution2D):
					W = np.asarray(f[keras_key][keras_key + "/kernel:0"])
					W = W.transpose(3,2,0,1)

					_assign(name, link.W, W)

				elif isinstance(link, L.Linear):
					W = np.asarray(f[keras_key][keras_key + "/kernel:0"])
					b = np.asarray(f[keras_key][keras_key + "/bias:0"])

					_assign(name, link.W, W.transpose(1,0))
					_assign(name, link.b, b)

				elif isinstance(link, L.BatchNormalization):
					beta = np.asarray(f[keras_key][keras_key + "/beta:0"])
					avg_mean = np.asarray(f[keras_key][keras_key + "/moving_mean:0"])
					avg_var = np.asarray(f[keras_key][keras_key + "/moving_variance:0"])

					_assign_batch_norm(name, link, beta, avg_mean, avg_var)
				else:
					raise ValueError("Unkown link type: {}!".format(type(link)))



from cvmodelz.models.meta_info import ModelInfo
from cvmodelz.models.wrapper import ModelWrapper
from cvmodelz.models.pretrained import PretrainedModelMixin


__all__ = [
	"ModelInfo",
	"ModelWrapper",
	"PretrainedModelMixin",
]



#!/usr/bin/env python
from pathlib import Path

from setuptools import setup, find_packages

try: # for pip >= 10
	from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
	from pip.req import parse_requirements
pkg_name = "cvmodelz"

cwd = Path(__file__).parent.resolve()
# Get __version__ variable
exec(open(str(cwd / pkg_name / '_version.py')).read())

install_requires = [
	line.strip()
		for line in open(str(cwd / "requirements.txt")).readlines()
]

setup(
	name=pkg_name,
	version=__version__,
	python_requires=">3.6",
	description='Wrapper for various computer vision models (mostly provided by chainer, chainercv, and chainercv2)',
	log_description=open(str(cwd / "README.md")).read(),
	author='Dimitri Korsch',
	author_email='korschdima@gmail.com',
	license='MIT License',
	packages=find_packages(exclude=("tests",)),
	zip_safe=False,
	setup_requires=[],
	install_requires=install_requires,
	package_data={'': ['requirements.txt']},
	data_files=[('.',['requirements.txt'])],
	include_package_data=True,
)



from cvmodelz.classifiers.base import Classifier
from cvmodelz.classifiers.separate_model_classifier import MeanModelClassifier
from cvmodelz.classifiers.separate_model_classifier import SeparateModelClassifier


__all__ = [
	"Classifier",
	"MeanModelClassifier",
	"SeparateModelClassifier",
]



import abc
from chainer import links as L

from cvmodelz.models.base import BaseModel
from cvmodelz.models.meta_info import ModelInfo

class PretrainedModelMixin(BaseModel):
	"""
		This mixin is designed to be a superclass besides one of
		chainer's built in models (VGG, ResNet, GoogLeNet).

		Example:

			import chainer.links as L
			class OurResNet(PretrainedModelMixin, L.ResNet50layers):
				...
	"""

	def __init__(self, *args, n_classes: int = 1000, pretrained_model: str = None, **kwargs):
		super().__init__(*args, pretrained_model=pretrained_model, **kwargs)

		with self.init_scope():
			self.init_extra_layers(n_classes)

		self.load(pretrained_model, strict=True)

	def forward(self, X, layer_name=None):
		assert hasattr(self, "meta"), "Did you forgot to initialize the meta attribute?"

		layer_name = layer_name or self.meta.classifier_layers[-1]
		caller = super().forward
		activations = caller(X, layers=[layer_name])

		if isinstance(activations, dict):
			activations = activations[layer_name]

		return activations

	def init_extra_layers(self, n_classes, **kwargs) -> None:
		if hasattr(self, self.clf_layer_name):
			delattr(self, self.clf_layer_name)

		clf_layer = L.Linear(self.meta.feature_size, n_classes)
		setattr(self, self.clf_layer_name, clf_layer)

	@property
	def model_instance(self) -> BaseModel:
		""" since it is a mixin, we are the model """

		return self




import chainer

from chainer import functions as F
from chainer import links as L
from chainer.links.model.vision.resnet import BuildingBlock
from chainer.links.model.vision.resnet import prepare
from collections import OrderedDict
from functools import partial


from cvmodelz.models.meta_info import ModelInfo
from cvmodelz.models.pretrained.base import PretrainedModelMixin

class BaseResNet(PretrainedModelMixin):
	n_layers = ""

	def init_model_info(self):
		self.meta = ModelInfo(
			name=f"ResNet{self.n_layers}",
			input_size=224,
			feature_size=2048,
			n_conv_maps=2048,

			conv_map_layer="res5",
			feature_layer="pool5",

			classifier_layers=["fc6"],

			prepare_func=prepare,
		)

	@property
	def functions(self):
		return super().functions

"""
We need this to "extract" pretrained_model argument,
otherwise it would be passed to the constructor of the
chainer.Chain class, where it raises an error
"""
class ResNet35Layers(chainer.Chain):

	def __init__(self, *args, pretrained_model=None, **kwargs):
		super().__init__(*args, **kwargs)


class ResNet35(BaseResNet, ResNet35Layers):
	n_layers = 35

	def init_extra_layers(self, *args, **kwargs):
		self.conv1 = L.Convolution2D(3, 64, 7, 2, 3, **kwargs)
		self.bn1 = L.BatchNormalization(64)
		self.res2 = BuildingBlock(2, 64, 64, 256, 1, **kwargs)
		self.res3 = BuildingBlock(3, 256, 128, 512, 2, **kwargs)
		self.res4 = BuildingBlock(3, 512, 256, 1024, 2, **kwargs)
		self.res5 = BuildingBlock(3, 1024, 512, 2048, 2, **kwargs)

		# the final fc layer is initilized by PretrainedModelMixin
		super().init_extra_layers(*args, **kwargs)

	@property
	def functions(self):
		links = [
			("conv1", [self.conv1, self.bn1, F.relu]),
			("pool1", [partial(F.max_pooling_2d, ksize=3, stride=2)]),
			("res2", [self.res2]),
			("res3", [self.res3]),
			("res4", [self.res4]),
			("res5", [self.res5]),
			("pool5", [self.pool]),
			("fc6", [self.fc6]),
			("prob", [F.softmax]),
		]
		return OrderedDict(links)

	def forward(self, x, layer_name=None):
		for key, funcs in self.functions.items():
			for func in funcs:
				x = func(x)
			if key == layer_name:
				return x


class ResNet50(BaseResNet, L.ResNet50Layers):
	n_layers = 50


class ResNet101(BaseResNet, L.ResNet101Layers):
	n_layers = 101


class ResNet152(BaseResNet, L.ResNet152Layers):
	n_layers = 152

